import argparse
import os
from typing import List

from wai.logging import LOGGING_WARNING

from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter
from idc.api import ImageSegmentationData, SplittableStreamWriter, make_list, AnnotationsOnlyWriter, \
    add_annotations_only_param, save_image, to_indexedpng
from simple_palette_utils import generate_palette_list, PALETTE_AUTO, palettes

DEFAULT_FILE_LIST = "data.txt"

DEFAULT_LABELS_LIST = "labels.txt"

DEFAULT_IMAGES_RELATIVE_PATH = "img"

DEFAULT_ANNOTATIONS_RELATIVE_PATH = "ann"


class PaddleImageSegmentationWriter(SplittableStreamWriter, AnnotationsOnlyWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None, files: str = None, labels: str = None, img_relative_path: str = None,
                 ann_relative_path: str = None, palette: str = None, annotations_only: bool = None, separator: str = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param output_dir: the output directory to save the data in
        :type output_dir: str
        :param files: the name of the file (no path) to store the images/annotations association in, uses DEFAULT_FILE_LIST if missing
        :type files: str
        :param labels: the name of the file (no path) to store the labels list in, uses DEFAULT_LABELS_LIST if missing
        :type labels: str
        :param img_relative_path: the relative path to use for the images, uses DEFAULT_IMAGES_RELATIVE_PATH if missing
        :type img_relative_path: str
        :param ann_relative_path: the relative path to use for the annotations, uses DEFAULT_ANNOTATIONS_RELATIVE_PATH if missing
        :type ann_relative_path: str
        :param palette: the palette to use, either a supported palette name (auto|x11|light|dark) or comma-separated list of R,G,B values
        :type palette: str
        :param annotations_only: whether to output only the annotations
        :type annotations_only: bool
        :param separator: the separator to use in the txt file (default is space)
        :type separator: str
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.files = files
        self.labels = labels
        self.img_relative_path = img_relative_path
        self.ann_relative_path = ann_relative_path
        self.palette = palette
        self.annotations_only = annotations_only
        self.separator = separator
        self._files = None
        self._palette_list = None
        self._labels = None
        self._output_dir_created = False

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-paddle-is"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the image segmentation in the specified text files listing image (with relative path) and annotation (with relative path). The labels get stored in a separate text file (one per line). When splitting, the split names get appended to the '--files' name (before the extension) separated by '-'."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the data in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        parser.add_argument("-f", "--files", metavar="NAME", type=str, default=DEFAULT_FILE_LIST, help="The text file to store the relation of images with their label indices in, e.g., 'data.txt'", required=False)
        parser.add_argument("-i", "--img_relative_path", metavar="PATH", type=str, default=DEFAULT_IMAGES_RELATIVE_PATH, help="The relative path to store the images under, e.g., 'img'", required=False)
        parser.add_argument("-a", "--ann_relative_path", metavar="PATH", type=str, default=DEFAULT_ANNOTATIONS_RELATIVE_PATH, help="The relative path to store the annotations under, e.g., 'ann'", required=False)
        parser.add_argument("-p", "--palette", metavar="PALETTE", type=str, default=PALETTE_AUTO, help="The palette to use; either palette name (%s) or comma-separated list of R,G,B values." % "|".join(palettes()), required=False)
        parser.add_argument("--labels", metavar="NAME", type=str, default=DEFAULT_LABELS_LIST, help="The name of the labels text file (no path), e.g., 'labels.txt'.", required=False)
        parser.add_argument("--separator", metavar="SEP", type=str, default=' ', help="The separator to use for reading the text files.", required=False)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.files = ns.files
        self.labels = ns.labels
        self.img_relative_path = ns.img_relative_path
        self.ann_relative_path = ns.ann_relative_path
        self.palette = ns.palette
        self.separator = ns.separator
        self.annotations_only = ns.annotations_only

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageSegmentationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._output_dir_created = False
        if self.files is None:
            self.files = DEFAULT_FILE_LIST
        if self.labels is None:
            self.labels = DEFAULT_LABELS_LIST
        if self.img_relative_path is None:
            self.img_relative_path = DEFAULT_IMAGES_RELATIVE_PATH
        if self.ann_relative_path is None:
            self.ann_relative_path = DEFAULT_ANNOTATIONS_RELATIVE_PATH
        self._palette_list = generate_palette_list(self.palette)
        if self.annotations_only is None:
            self.annotations_only = False
        if self.separator is None:
            self.separator = ' '
        self._files = dict()
        self._labels = list()

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        output_dir = self.session.expand_placeholders(self.output_dir)
        if (not self._output_dir_created) and (not os.path.exists(output_dir)):
            self._output_dir_created = True
            self.logger().info("Creating output dir: %s" % output_dir)
            os.makedirs(output_dir)
        for item in make_list(data):
            split = ''
            if self.splitter is not None:
                split = self.splitter.next(item=item.image_name)
            if not self.annotations_only:
                path = os.path.join(output_dir, self.img_relative_path)
                if not os.path.exists(path):
                    self.logger().info("Creating image dir: %s" % path)
                    os.makedirs(path)
            if split not in self._files:
                self._files[split] = list()

            # records labels
            if len(self._labels) == 0:
                self._labels.extend(item.annotation.labels)

            # relative file names
            img_relative_name = os.path.join(self.img_relative_path, item.image_name)
            ann_relative_name = os.path.join(self.ann_relative_path, os.path.splitext(item.image_name)[0] + ".png")

            # is separator part of file name?
            if self.separator in img_relative_name:
                raise Exception("Found separator '%s' in image name '%s'! This would create an unparsable dataset, use a custom separator instead!" % (self.separator, img_relative_name))
            if self.separator in ann_relative_name:
                raise Exception("Found separator '%s' in image name '%s'! This would create an unparsable dataset, use a custom separator instead!" % (self.separator, ann_relative_name))

            # update file map
            self._files[split].append(img_relative_name + self.separator + ann_relative_name)

            path = os.path.join(output_dir, img_relative_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)
            path = os.path.join(output_dir, ann_relative_name)
            ann = to_indexedpng(item.image_width, item.image_height, item.annotation, self._palette_list, background=0)
            self.logger().info("Writing annotations to: %s" % path)
            save_image(ann, path, make_dirs=True)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        first = True
        output_dir = self.session.expand_placeholders(self.output_dir)
        for sub_set in self._files:
            # write label list
            if first:
                first = False
                path = os.path.join(output_dir, self.labels)
                self.logger().info("Writing label list: %s" % path)
                with open(path, "w") as fp:
                    for label in self._labels:
                        fp.write(label)
                        fp.write("\n")

            # write file list
            if len(sub_set) > 0:
                path = os.path.join(output_dir, os.path.splitext(self.files)[0] + "-" + sub_set + os.path.splitext(self.files)[1])
            else:
                path = os.path.join(output_dir, self.files)
            self.logger().info("Writing file list: %s" % path)
            with open(path, "w") as fp:
                for line in self._files[sub_set]:
                    fp.write(line)
                    fp.write("\n")



import argparse
import os.path
from typing import List, Iterable, Union

from seppl.placeholders import PlaceholderSupporter, placeholder_list
from seppl.io import locate_files
from wai.logging import LOGGING_WARNING

from idc.api import ImageClassificationData
from idc.api import Reader


class PaddleImageClassificationReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 id_label_map: str = None, resume_from: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param id_label_map: the file with the mapping of label ID and label string
        :type id_label_map: str
        :param resume_from: the file to resume from (glob)
        :type resume_from: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.id_label_map = id_label_map
        self.resume_from = resume_from
        self._inputs = None
        self._current_input = None
        self._id_label_map = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-paddle-ic"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the image classification from the specified text files listing image (with relative path) and the associate label ID. The label ID to text mapping can be supplied as separate text file (ID <space> text)."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the text file(s) to read; glob syntax is supported; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the text files to use; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("--resume_from", type=str, help="Glob expression matching the file to resume from, e.g., '*/012345.txt'", required=False)
        parser.add_argument("-m", "--id_label_map", metavar="FILE", type=str, default=None, help="The mapping between label ID and text (ID <space> text); " + placeholder_list(obj=self), required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.id_label_map = ns.id_label_map
        self.resume_from = ns.resume_from

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageClassificationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        self._id_label_map = None
        if self.id_label_map is not None:
            self._id_label_map = dict()
            id_label_map = self.session.expand_placeholders(self.id_label_map)
            self.logger().info("Loading ID/label map: %s" % id_label_map)
            with (open(id_label_map) as fp):
                for line in fp.readlines():
                    line = line.strip()
                    if (len(line) > 0) and (" " in line):
                        try:
                            label_id = str(int(line[0:line.index(" ")].strip()))
                            label_text = line[line.index(" ") + 1:].strip()
                            self._id_label_map[label_id] = label_text
                        except:
                            self.logger().warning("Failed to parse: %s" % line)

        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.txt", resume_from=self.resume_from)

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))
        current_dir = os.path.dirname(self.session.current_input)
        with open(self.session.current_input) as fp:
            for line in fp.readlines():
                line = line.strip()
                if len(line) == 0:
                    continue
                if " " in line:
                    try:
                        label = str(int(line[line.rindex(" ")+1:].strip()))
                        if label in self._id_label_map:
                            label = self._id_label_map[label]
                        fname = os.path.join(current_dir, line[0:line.rindex(" ")].strip())
                        yield ImageClassificationData(source=fname, annotation=label)
                    except:
                        self.logger().warning("Failed to parse: %s" % line)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



import argparse
import os.path
from typing import List, Iterable, Union

from seppl.placeholders import PlaceholderSupporter, placeholder_list
from seppl.io import locate_files
from wai.logging import LOGGING_WARNING

from idc.api import ImageSegmentationData, load_image_from_file, from_indexedpng
from idc.api import Reader


class PaddleImageSegmentationReader(Reader, PlaceholderSupporter):

    def __init__(self, source: Union[str, List[str]] = None, source_list: Union[str, List[str]] = None,
                 labels_file: str = None, labels: List[str] = None, separator: str = None, resume_from: str = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param source: the filename(s)
        :param source_list: the file(s) with filename(s)
        :param labels_file: the file with the labels (incl background), one per line
        :type labels_file: str
        :param labels: the list of labels to use
        :type labels: list
        :param resume_from: the file to resume from (glob)
        :type resume_from: str
        :param separator: the separator to use in the txt file (default is space)
        :type separator: str
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(logger_name=logger_name, logging_level=logging_level)
        self.source = source
        self.source_list = source_list
        self.labels_file = labels_file
        self.labels = labels
        self.separator = separator
        self.resume_from = resume_from
        self._inputs = None
        self._current_input = None
        self._labels = None
        self._label_mapping = None

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "from-paddle-is"

    def description(self) -> str:
        """
        Returns a description of the reader.

        :return: the description
        :rtype: str
        """
        return "Loads the image segmentation from the specified text files listing image (JPG with relative path) and annotation image (indexed PNG with relative path). The labels (incl background) are specified via a separate text file."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-i", "--input", type=str, help="Path to the text file(s) to read; glob syntax is supported; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("-I", "--input_list", type=str, help="Path to the text file(s) listing the text files to use; " + placeholder_list(obj=self), required=False, nargs="*")
        parser.add_argument("--resume_from", type=str, help="Glob expression matching the file to resume from, e.g., '*/012345.txt'", required=False)
        parser.add_argument("--labels_file", metavar="FILE", type=str, default=None, help="The file with the labels associated with the indices (incl. background); " + placeholder_list(obj=self), required=False)
        parser.add_argument("--labels", metavar="LABEL", type=str, default=None, help="The labels that the indices represent (incl background).", nargs="*")
        parser.add_argument("--separator", metavar="SEP", type=str, default=' ', help="The separator to use for reading the text files.", required=False)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.source = ns.input
        self.source_list = ns.input_list
        self.labels_file = ns.labels_file
        self.labels = ns.labels
        self.resume_from = ns.resume_from
        self.separator = ns.separator

    def generates(self) -> List:
        """
        Returns the list of classes that get produced.

        :return: the list of classes
        :rtype: list
        """
        return [ImageSegmentationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()

        self._label_mapping = None
        self._labels = None
        if self.labels_file is not None:
            self._labels = list()
            self._label_mapping = dict()
            labels_file = self.session.expand_placeholders(self.labels_file)
            self.logger().info("Loading labels: %s" % labels_file)
            with (open(labels_file) as fp):
                for line in fp.readlines():
                    line = line.strip()
                    if len(line) > 0:
                        self._labels.append(line)
                        self._label_mapping[len(self._label_mapping)] = line
        elif self.labels is not None:
            self._labels = list()
            self._label_mapping = dict()
            for label in self.labels:
                self._labels.append(label)
                self._label_mapping[len(self._label_mapping)] = label
        if self._labels is None:
            raise Exception("Neither labels file nor explicit labels list specified!")
        if len(self._labels) == 0:
            raise Exception("Empty list of labels (background must be listed explicitly)!")
        self.logger().info("# labels: %d" % len(self._labels))
        self.logger().debug("label mapping: %s" % str(self._label_mapping))

        if self.separator is None:
            self.separator = ' '

        self._inputs = locate_files(self.source, input_lists=self.source_list, fail_if_empty=True, default_glob="*.txt", resume_from=self.resume_from)

    def read(self) -> Iterable:
        """
        Loads the data and returns the items one by one.

        :return: the data
        :rtype: Iterable
        """
        self._current_input = self._inputs.pop(0)
        self.session.current_input = self._current_input
        self.logger().info("Reading from: " + str(self.session.current_input))
        current_dir = os.path.dirname(self.session.current_input)
        with open(self.session.current_input) as fp:
            for line in fp.readlines():
                line = line.strip()
                if len(line) == 0:
                    continue
                if self.separator in line:
                    try:
                        img_file = os.path.join(current_dir, line[0:line.rindex(self.separator)].strip())
                        ann_file = os.path.join(current_dir, line[line.rindex(self.separator)+1:].strip())
                        ann = load_image_from_file(ann_file)
                        annotations = from_indexedpng(ann, self._labels, self._label_mapping, self.logger())
                        yield ImageSegmentationData(source=img_file, annotation=annotations)
                    except:
                        self.logger().warning("Failed to parse: %s" % line)

    def has_finished(self) -> bool:
        """
        Returns whether reading has finished.

        :return: True if finished
        :rtype: bool
        """
        return len(self._inputs) == 0



from typing import List, Dict


def list_classes() -> Dict[str, List[str]]:
    return {
        "seppl.io.Reader": [
            "idc.paddle.reader.imgcls",
            "idc.paddle.reader.imgseg",
        ],
        "seppl.io.Writer": [
            "idc.paddle.writer.imgcls",
            "idc.paddle.writer.imgseg",
        ],
    }



import argparse
import os
from typing import List

from wai.logging import LOGGING_WARNING

from seppl.placeholders import placeholder_list, InputBasedPlaceholderSupporter
from idc.api import ImageClassificationData, SplittableStreamWriter, make_list, AnnotationsOnlyWriter, \
    add_annotations_only_param

DEFAULT_ID_LABEL_MAP = "labels.map"

DEFAULT_FILE_LABEL_MAP = "annotations.txt"

DEFAULT_RELATIVE_PATH = "images"


class PaddleImageClassificationWriter(SplittableStreamWriter, AnnotationsOnlyWriter, InputBasedPlaceholderSupporter):

    def __init__(self, output_dir: str = None, file_label_map: str = None, relative_path: str = None,
                 id_label_map: str = None, annotations_only: bool = None,
                 split_names: List[str] = None, split_ratios: List[int] = None,
                 logger_name: str = None, logging_level: str = LOGGING_WARNING):
        """
        Initializes the reader.

        :param output_dir: the output directory to save the data in
        :type output_dir: str
        :param file_label_map: the name of the file (no path) to store the image file name/label index relation in, uses DEFAULT_FILE_LABEL_MAP if missing
        :type file_label_map: str
        :param relative_path: the relative path to use for the images, uses DEFAULT_RELATIVE_PATH if missing
        :type relative_path: str
        :param id_label_map: the name of the file (no path) to store the label index/label string relation in, uses DEFAULT_ID_LABEL_MAP if missing
        :type id_label_map: str
        :param split_names: the names of the splits, no splitting if None
        :type split_names: list
        :param split_ratios: the integer ratios of the splits (must sum up to 100)
        :type split_ratios: list
        :param logger_name: the name to use for the logger
        :type logger_name: str
        :param logging_level: the logging level to use
        :type logging_level: str
        """
        super().__init__(split_names=split_names, split_ratios=split_ratios, logger_name=logger_name, logging_level=logging_level)
        self.output_dir = output_dir
        self.file_label_map = file_label_map
        self.relative_path = relative_path
        self.id_label_map = id_label_map
        self.annotations_only = annotations_only
        self._file_label_maps = None
        self._id_label_map = None
        self._output_dir_created = False

    def name(self) -> str:
        """
        Returns the name of the handler, used as sub-command.

        :return: the name
        :rtype: str
        """
        return "to-paddle-ic"

    def description(self) -> str:
        """
        Returns a description of the writer.

        :return: the description
        :rtype: str
        """
        return "Saves the image classification in the specified text files listing image (with relative path) and the associate label ID. The label ID to text mapping is stored in a separate text file (ID <space> text)."

    def _create_argparser(self) -> argparse.ArgumentParser:
        """
        Creates an argument parser. Derived classes need to fill in the options.

        :return: the parser
        :rtype: argparse.ArgumentParser
        """
        parser = super()._create_argparser()
        parser.add_argument("-o", "--output", type=str, help="The directory to store the data in. Any defined splits get added beneath there. " + placeholder_list(obj=self), required=True)
        parser.add_argument("-f", "--file_label_map", metavar="NAME", type=str, default=DEFAULT_FILE_LABEL_MAP, help="The text file to store the relation of images with their label indices in, e.g., 'annotations.txt'", required=False)
        parser.add_argument("-p", "--relative_path", metavar="PATH", type=str, default=DEFAULT_RELATIVE_PATH, help="The relative path to the annotations text file to store the images under, e.g., 'images'", required=False)
        parser.add_argument("-m", "--id_label_map", metavar="NAME", type=str, default=DEFAULT_ID_LABEL_MAP, help="The name of the ID/label text mapping text file (no path), e.g., 'labels.map'.", required=False)
        add_annotations_only_param(parser)
        return parser

    def _apply_args(self, ns: argparse.Namespace):
        """
        Initializes the object with the arguments of the parsed namespace.

        :param ns: the parsed arguments
        :type ns: argparse.Namespace
        """
        super()._apply_args(ns)
        self.output_dir = ns.output
        self.file_label_map = ns.file_label_map
        self.id_label_map = ns.id_label_map
        self.relative_path = ns.relative_path
        self.annotations_only = ns.annotations_only

    def accepts(self) -> List:
        """
        Returns the list of classes that are accepted.

        :return: the list of classes
        :rtype: list
        """
        return [ImageClassificationData]

    def initialize(self):
        """
        Initializes the processing, e.g., for opening files or databases.
        """
        super().initialize()
        self._output_dir_created = False
        if self.file_label_map is None:
            self.file_label_map = DEFAULT_FILE_LABEL_MAP
        if self.id_label_map is None:
            self.id_label_map = DEFAULT_ID_LABEL_MAP
        if self.relative_path is None:
            self.relative_path = DEFAULT_RELATIVE_PATH
        if self.annotations_only is None:
            self.annotations_only = False
        self._file_label_maps = dict()
        self._id_label_map = dict()

    def write_stream(self, data):
        """
        Saves the data one by one.

        :param data: the data to write (single record or iterable of records)
        """
        output_dir = self.session.expand_placeholders(self.output_dir)
        if (not self._output_dir_created) and (not os.path.exists(output_dir)):
            self._output_dir_created = True
            self.logger().info("Creating output dir: %s" % output_dir)
            os.makedirs(output_dir)
        for item in make_list(data):
            sub_dir = output_dir
            if self.splitter is not None:
                split = self.splitter.next(item=item.image_name)
                sub_dir = os.path.join(sub_dir, split)
            if not os.path.exists(sub_dir):
                self.logger().info("Creating dir: %s" % sub_dir)
                os.makedirs(sub_dir)
            if not self.annotations_only:
                path = os.path.join(sub_dir, self.relative_path)
                if not os.path.exists(path):
                    self.logger().info("Creating image dir: %s" % path)
                    os.makedirs(path)
            if sub_dir not in self._file_label_maps:
                self._file_label_maps[sub_dir] = dict()

            # relative file name
            relative_name = os.path.join(self.relative_path, item.image_name)

            # space in file name?
            if " " in relative_name:
                raise Exception("Image name cannot contain space(s): %s" % relative_name)

            # update label map
            if item.has_annotation():
                if item.annotation not in self._id_label_map:
                    self._id_label_map[item.annotation] = len(self._id_label_map)

            # update file map
            self._file_label_maps[sub_dir][relative_name] = self._id_label_map[item.annotation]

            path = os.path.join(sub_dir, self.relative_path, item.image_name)
            if not self.annotations_only:
                self.logger().info("Writing image to: %s" % path)
                item.save_image(path)

    def finalize(self):
        """
        Finishes the processing, e.g., for closing files or databases.
        """
        super().finalize()
        for sub_dir in self._file_label_maps:
            # dict with id as key
            swapped = dict((v, k) for k, v in self._id_label_map.items())
            # write label map
            path = os.path.join(sub_dir, self.id_label_map)
            self.logger().info("Writing label map: %s" % path)
            with open(path, "w") as fp:
                for key in sorted(swapped.keys()):
                    fp.write("%d %s\n" % (key, swapped[key]))

            # write file map
            path = os.path.join(sub_dir, self.file_label_map)
            self.logger().info("Writing file map: %s" % path)
            with open(path, "w") as fp:
                for key in sorted(self._file_label_maps[sub_dir].keys()):
                    fp.write("%s %d\n" % (key, self._file_label_maps[sub_dir][key]))
